import os
import sys
import unittest

try:
    from unqlite import UnQLite
    from unqlite.core import CursorIterator
except ImportError:
    sys.stderr.write('Unable to import `unqlite`. Make sure it is properly '
                     'installed.\n')
    sys.stderr.flush()
    raise

from _unqlite import UNQLITE_OK


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.db = UnQLite(':mem:')

    def store_range(self, n):
        for i in range(n):
            self.db['k%s' % i] = str(i)


class TestKeyValueStorage(BaseTestCase):
    def test_basic_operations(self):
        self.db.store('k1', 'v1')
        self.db.store('k2', 'v2')
        self.assertEqual(self.db.fetch('k1'), 'v1')
        self.assertEqual(self.db.fetch('k2'), 'v2')
        self.assertRaises(KeyError, self.db.fetch, 'k3')

        self.db.delete('k2')
        self.assertRaises(KeyError, self.db.fetch, 'k2')

        self.assertTrue(self.db.exists('k1'))
        self.assertFalse(self.db.exists('k2'))

    def test_dict_interface(self):
        self.db['k1'] = 'v1'
        self.db['k2'] = 'v2'
        self.assertEqual(self.db['k1'], 'v1')
        self.assertEqual(self.db['k2'], 'v2')
        self.assertRaises(KeyError, lambda: self.db['k3'])

        del self.db['k2']
        self.assertRaises(KeyError, lambda: self.db['k2'])

        self.assertTrue('k1' in self.db)
        self.assertFalse('k2' in self.db)

    def test_append(self):
        self.db['k1'] = 'v1'
        self.db.append('k1', 'V1')
        self.assertEqual(self.db['k1'], 'v1V1')

        self.db.append('k2', 'V2')
        self.assertEqual(self.db['k2'], 'V2')

    def test_store_fmt(self):
        self.db.store_fmt('k1', 'foo %s:%d:%s', 'v1', 25, 'VX')
        self.assertEqual(self.db['k1'], 'foo v1:25:VX')

    def test_fetch_cb(self):
        state = []

        def cb(data):
            state.append(data)

        def alt_cb(data):
            state.append(data.upper())

        self.db['k1'] = 'v1'
        self.db['k2'] = 'v2'

        self.db.fetch_cb('k1', cb)
        self.assertEqual(state, ['v1'])

        self.db.fetch_cb('k2', cb)
        self.assertEqual(state, ['v1', 'v2'])

        self.db.fetch_cb('k1', alt_cb)
        self.assertEqual(state, ['v1', 'v2', 'V1'])

        self.assertRaises(KeyError, lambda: self.db.fetch_cb('kx', cb))

    def test_iteration(self):
        self.store_range(4)
        data = [item for item in self.db]
        self.assertEqual(data, [
            ('k0', '0'),
            ('k1', '1'),
            ('k2', '2'),
            ('k3', '3'),
        ])

        del self.db['k2']
        self.assertEqual([key for key, _ in self.db], ['k0', 'k1', 'k3'])

    def test_range(self):
        self.store_range(10)
        data = [item for item in self.db.range('k4', 'k6')]
        self.assertEqual(data, [
            ('k4', '4'),
            ('k5', '5'),
            ('k6', '6'),
        ])

        data = [item for item in self.db.range('k8', 'kX')]
        self.assertEqual(data, [
            ('k8', '8'),
            ('k9', '9'),
        ])

        def invalid_start():
            data = [item for item in self.db.range('kx', 'k2')]
        self.assertRaises(Exception, invalid_start)


class TestTransaction(BaseTestCase):
    """
    We must use a file-based database to test the transaction functions. See
    http://unqlite.org/forum/trouble-with-transactions+1 for details.
    """
    def setUp(self):
        self._filename = 'test.db'
        self.db = UnQLite(self._filename)

    def tearDown(self):
        try:
            self.db.close()
        except:
            pass
        if os.path.exists(self._filename):
            os.unlink(self._filename)

    def test_transaction(self):
        @self.db.commit_on_success
        def _test_success(key, value):
            self.db[key] = value

        @self.db.commit_on_success
        def _test_failure(key, value):
            self.db[key] = value
            raise Exception('intentional exception raised')

        _test_success('k1', 'v1')
        self.assertEqual(self.db['k1'], 'v1')

        self.assertRaises(Exception , lambda: _test_failure('k2', 'v2'))
        self.assertRaises(KeyError, lambda: self.db['k2'])

    def test_explicit_transaction(self):
        self.db.close()
        self.db.open()
        self.db.begin()
        self.db['k1'] = 'v1'
        self.db.rollback()

        self.assertRaises(KeyError, lambda: self.db['k1'])


class TestCursor(BaseTestCase):
    def setUp(self):
        super(TestCursor, self).setUp()
        self.store_range(10)

    def assertIndex(self, cursor, idx):
        self.assertTrue(cursor.is_valid())
        self.assertEqual(cursor.key(), 'k%d' % idx)
        self.assertEqual(cursor.value(), str(idx))

    def test_cursor_basic(self):
        cursor = self.db.cursor()
        self.assertIndex(cursor, 0)
        cursor.next()
        self.assertIndex(cursor, 1)
        cursor.last()
        self.assertIndex(cursor, 9)
        cursor.previous()
        self.assertIndex(cursor, 8)
        cursor.first()
        self.assertIndex(cursor, 0)
        cursor.delete()
        self.assertIndex(cursor, 1)
        cursor.close()

    def test_cursor_iteration(self):
        with self.db.cursor() as cursor:
            cursor.seek('k4')
            cursor.delete()
            cursor.reset()
            results = [item for item in cursor]
            self.assertEqual(results, [
                ('k0', '0'),
                ('k1', '1'),
                ('k2', '2'),
                ('k3', '3'),
                ('k5', '5'),
                ('k6', '6'),
                ('k7', '7'),
                ('k8', '8'),
                ('k9', '9'),
            ])

            cursor.seek('k5')
            self.assertEqual(cursor.value(), '5')
            keys = [key for key, _ in cursor]
            self.assertEqual(keys, ['k5', 'k6', 'k7', 'k8', 'k9'])

        with self.db.cursor() as cursor:
            self.assertRaises(Exception, cursor.seek, 'k4')
            cursor.seek('k5')
            keys = []
            while True:
                key = cursor.key()
                keys.append(key)
                if key == 'k7':
                    break
                else:
                    cursor.next()
        self.assertEqual(keys, ['k5', 'k6', 'k7'])

    def test_iterate_count(self):
        with self.db.cursor() as cursor:
            cursor_i = CursorIterator(cursor, 3)
            items = [item for item in cursor_i]
            self.assertEqual(items, [
                ('k0', '0'),
                ('k1', '1'),
                ('k2', '2'),
            ])

        with self.db.cursor() as cursor:
            cursor.next()
            items = [item for item in cursor.fetch_count(2)]
            self.assertEqual(items, [
                ('k1', '1'),
                ('k2', '2'),
            ])

        with self.db.cursor() as cursor:
            cursor.seek('k3')
            items = [item for item in cursor.fetch_until('k6')]
            self.assertEqual(items, [
                ('k3', '3'),
                ('k4', '4'),
                ('k5', '5'),
                ('k6', '6'),
            ])

            cursor.seek('k1')
            items = [item for item in cursor.fetch_until('k4', False)]
            self.assertEqual(items, [
                ('k1', '1'),
                ('k2', '2'),
                ('k3', '3'),
            ])

    def test_cursor_callbacks(self):
        keys = []
        values = []
        with self.db.cursor() as cursor:
            cursor.last()

            @cursor.key_callback
            def kcb(key):
                keys.append(key)

            def vcb(value):
                values.append(value)
            cursor.value_callback(vcb)

            self.assertEqual(keys, ['k9'])
            self.assertEqual(values, ['9'])

            cursor.previous()
            cursor.value_callback(vcb)

            self.assertEqual(keys, ['k9'])
            self.assertEqual(values, ['9', '8'])


class TestJx9(BaseTestCase):
    def test_simple_compilation(self):
        script = """
            $collection = 'users';
            if (!db_exists($collection)) {
                db_create($collection);
            }
            db_store($collection, {"username": "huey", "age": 3});
            $huey_id = db_last_record_id($collection);
            db_store($collection, {"username": "mickey", "age": 5});
            $mickey_id = db_last_record_id($collection);
            $something = 'hello world';
            $users = db_fetch_all($collection);
            $nested = {
                "k1": {"foo": [1, 2, 3]},
                "k2": ["v2", ["v3", "v4"]]};
        """

        with self.db.compile_script(script) as vm:
            vm.execute()
            self.assertEqual(vm['huey_id'], 0)
            self.assertEqual(vm['mickey_id'], 1)
            self.assertEqual(vm['something'], 'hello world')

            users = vm['users']
            self.assertEqual(users, [
                {'__id': 0, 'age': 3, 'username': 'huey'},
                {'__id': 1, 'age': 5, 'username': 'mickey'},
            ])

            nested = vm['nested']
            self.assertEqual(nested, {
                'k1': {'foo': [1, 2, 3]},
                'k2': ['v2', ['v3', 'v4']]})

    def test_setting_values(self):
        script = """
            $collection = 'users';
            db_create($collection);
            db_store($collection, $values);
            $users = db_fetch_all($collection);
        """
        values = [
            {'username': 'hubie', 'color': 'white'},
            {'username': 'michael', 'color': 'black'},
        ]

        with self.db.compile_script(script) as vm:
            vm['values'] = values
            vm.execute()

            users = vm['users']
            self.assertEqual(users, [
                {'username': 'hubie', 'color': 'white', '__id': 0},
                {'username': 'michael', 'color': 'black', '__id': 1},
            ])

    def test_foreign_function(self):
        script = "db_create('values'); db_store('values', $values);"
        values = [{'val': i} for i in range(20)]
        with self.db.compile_script(script) as vm:
            vm['values'] = values
            vm.execute()

        script = "$ret = db_fetch_all('values', my_filter_func);"
        with self.db.compile_script(script) as vm:
            @vm.foreign_function('my_filter_func')
            def _filter_func(context, obj):
                return obj['val'] in range(7, 13)

            vm.execute()
            result = vm['ret']

        self.assertEqual(result, [
            {'__id': 7, 'val': 7},
            {'__id': 8, 'val': 8},
            {'__id': 9, 'val': 9},
            {'__id': 10, 'val': 10},
            {'__id': 11, 'val': 11},
            {'__id': 12, 'val': 12},
        ])


class TestUtils(BaseTestCase):
    def test_random(self):
        ri = self.db.random_number()
        self.assertTrue(isinstance(ri, (int, long)))

        rs = self.db.random_string(10)
        self.assertEqual(len(rs), 10)

    def test_store_file(self):
        cur_dir = os.path.realpath(os.path.dirname(__file__))
        filename = os.path.join(cur_dir, 'core.py')
        with open(filename) as fh:
            contents = fh.read()

        self.db.store_file('source', filename)
        db_contents = self.db.fetch('source', 1024 * 64)
        self.assertEqual(db_contents, contents)


class TestCollection(BaseTestCase):
    def test_basic_operations(self):
        users = self.db.collection('users')
        self.assertFalse(users.exists())
        users.create()
        self.assertTrue(users.exists())
        self.assertEqual(len(users), 0)

        user_data = [
            {'name': 'charlie', 'activities': ['coding', 'reading']},
            {'name': 'huey', 'activities': ['playing', 'sleeping']},
            {'name': 'mickey', 'activities': ['sleeping', 'hunger']}]

        users.store(user_data)
        self.assertEqual(len(users), 3)

        users_with_ids = [record.copy() for record in user_data]
        for idx, record in enumerate(users_with_ids):
            record['__id'] = idx

        results = users.all()
        self.assertEqual(results, users_with_ids)

        users.store({'name': 'leslie', 'activities': ['reading', 'surgery']})
        self.assertEqual(len(users), 4)

        record = users.fetch_current()
        self.assertEqual(record['name'], 'charlie')

        self.assertEqual(users.fetch(3), {
            'name': 'leslie',
            'activities': ['reading', 'surgery'],
            '__id': 3})

        users.delete(0)
        users.delete(2)
        users.delete(3)
        self.assertEqual(users.all(), [
            {'name': 'huey', 'activities': ['playing', 'sleeping'], '__id': 1}
        ])

        self.assertIsNone(users[99])

    def test_filtering(self):
        values = self.db.collection('values')
        values.create()
        value_data = [{'val': i} for i in range(20)]
        values.store(value_data)
        self.assertEqual(len(values), 20)

        filtered = values.filter(lambda obj: obj['val'] in range(7, 12))
        self.assertEqual(filtered, [
            {'__id': 7, 'val': 7},
            {'__id': 8, 'val': 8},
            {'__id': 9, 'val': 9},
            {'__id': 10, 'val': 10},
            {'__id': 11, 'val': 11},
        ])


if __name__ == '__main__':
    unittest.main(argv=sys.argv)



from unqlite.core import UnQLite
from unqlite._unqlite import (
    UNQLITE_OPEN_READONLY,
    UNQLITE_OPEN_READWRITE,
    UNQLITE_OPEN_CREATE,
    UNQLITE_OPEN_EXCLUSIVE,
    UNQLITE_OPEN_TEMP_DB,
    UNQLITE_OPEN_NOMUTEX,
    UNQLITE_OPEN_OMIT_JOURNALING,
    UNQLITE_OPEN_IN_MEMORY,
    UNQLITE_OPEN_MMAP)



#!/usr/bin/env python
# -*- coding: ascii -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

"""
Ctypesgencore is the module that contains the main body of ctypesgen - in fact,
it contains everything but the command-line interface.

ctypesgen's job is divided into three steps:

Step 1: Parse

Ctypesgen reads the input header files and parses them. It generates a list of
function, variable, struct, union, enum, constant, typedef, and macro
descriptions from the input files. These descriptions are encapsulated as
ctypesgen.descriptions.Description objects.

The package ctypesgen.parser is responsible for the parsing stage.

Step 2: Process

Ctypesgen processes the list of descriptions from the parsing stage. This is
the stage where ctypesgen resolves name conflicts and filters descriptions using
the regexes specified on the command line. Other processing steps take place
at this stage, too. When processing is done, ctypesgen finalizes which
descriptions will be included in the output file.

The package ctypesgen.processor is responsible for the processing stage.

Step 3: Print

Ctypesgen writes the descriptions to the output file, along with a header.

The package ctypesgen.printer is responsible for the printing stage.

There are three modules in ctypesgencore that describe the format that the
parser, processor, and printer modules use to pass information. They are:

* descriptions: Classes to represent the descriptions.

* ctypedecls: Classes to represent C types.

* expressions: Classes to represent an expression in a language-independent
format.
"""

__version__ = '0.r125'
VERSION = __version__

__all__ = ["parser","processor","printer",
           "descriptions","ctypedescs","expressions",
           "messages","options"]

# Workhorse modules
import parser
import processor
import printer

# Modules describing internal format
import descriptions
import ctypedescs
import expressions

# Helper modules
import messages
import options



from contextlib import contextmanager
from functools import wraps

from unqlite._unqlite import *
from unqlite._unqlite import _libs as _c_libraries
from unqlite._unqlite import _variadic_function

DEFAULT_BUFFER_SIZE = 16384


def handle_return_value(rc):
    if rc != UNQLITE_OK:
        raise Exception({
            UNQLITE_NOTFOUND: 'Key not found',
            UNQLITE_NOMEM: 'Out of memory',
            UNQLITE_ABORT: 'Executed command request an operation abort',
            UNQLITE_IOERR: 'OS error',
            UNQLITE_CORRUPT: 'Corrupt database',
            UNQLITE_LOCKED: 'Locked database',
            UNQLITE_BUSY: 'Database is locked by another thread/process',
            UNQLITE_DONE: 'Error: UNQLITE_DONE',
            UNQLITE_NOTIMPLEMENTED: 'Not implemented',
            UNQLITE_READ_ONLY: 'Database is in read-only mode',
        }.get(rc, 'Unknown exception: %s' % rc))
    return True

def kv_callback(fn):
    def cb_wrapper(data_ptr, data_len, user_data_ptr):
        data = (c_char * data_len).from_address(data_ptr).raw
        try:
            fn(data)
        except:
            return UNQLITE_ABORT
        return UNQLITE_OK

    c_callback = CFUNCTYPE(
        UNCHECKED(c_int),
        POINTER(None),
        UNCHECKED(c_uint),
        POINTER(None))(cb_wrapper)

    fn._c_callback = c_callback
    fn._cb_wrapper = cb_wrapper

    return fn

def _value_to_string(ptr):
    nbytes = c_int()
    res = unqlite_value_to_string(ptr, pointer(nbytes))
    return res.raw[:nbytes.value]

def _array_fetch_cb(array_ptr, callback):
    c_callback = CFUNCTYPE(
        UNCHECKED(c_int),
        POINTER(unqlite_value),
        POINTER(unqlite_value),
        POINTER(None))(callback)
    unqlite_array_walk(array_ptr, c_callback, None)

def _value_to_list(ptr):
    accum = []
    def cb(key_ptr, value_ptr, user_data_ptr):
        accum.append(_convert_value(value_ptr))
        return UNQLITE_OK
    _array_fetch_cb(ptr, cb)
    return accum

def _value_to_dict(ptr):
    accum = {}
    def cb(key_ptr, value_ptr, user_data_ptr):
        accum[_convert_value(key_ptr)] = _convert_value(value_ptr)
        return UNQLITE_OK
    _array_fetch_cb(ptr, cb)
    return accum

def _convert_value(ptr):
    if unqlite_value_is_json_object(ptr):
        return _value_to_dict(ptr)
    elif unqlite_value_is_json_array(ptr):
        return _value_to_list(ptr)
    elif unqlite_value_is_string(ptr):
        return _value_to_string(ptr)
    elif unqlite_value_is_int(ptr):
        return unqlite_value_to_int(ptr)
    elif unqlite_value_is_float(ptr):
        return unqlite_value_to_float(ptr)
    elif unqlite_value_is_bool(ptr):
        return bool(unqlite_value_to_bool(ptr))
    elif unqlite_value_is_null(ptr):
        return None
    raise TypeError('Unrecognized type: %s' % ptr)

def wrap_foreign_function(fn):
    def inner(context, nargs, values):
        converted_args = [_convert_value(values[i]) for i in range(nargs)]
        context_wrapper = _Context(context)
        try:
            ret = fn(context_wrapper, *converted_args)
        except:
            return UNQLITE_ABORT
        else:
            context_wrapper.push_result(ret)
            return UNQLITE_OK

    c_callback = CFUNCTYPE(
        UNCHECKED(c_int),
        POINTER(unqlite_context),
        c_int,
        POINTER(POINTER(unqlite_value)))(inner)

    return c_callback, inner

_unqlite_lib = _c_libraries['unqlite']


class UnQLite(object):
    """
    UnQLite database python bindings.
    """
    def __init__(self, database=':mem:', open_manually=False):
        self.database = database
        self._unqlite = POINTER(unqlite)()
        self._command_registry = {}
        if not open_manually:
            self.open()

    def open(self, flags=UNQLITE_OPEN_CREATE):
        rc = unqlite_open(byref(self._unqlite), self.database, flags)
        if rc != UNQLITE_OK:
            raise Exception('Unable to open UnQLite database')

    def close(self):
        return handle_return_value(unqlite_close(self._unqlite))

    def config(self, verb, *args):
        return handle_return_value(_unqlite_lib.unqlite_config(
            self._unqlite,
            verb,
            *args))

    def _disable_autocommit(self):
        self.config(UNQLITE_CONFIG_DISABE_AUTO_COMMIT)

    def _set_max_page_cache(self, npages):
        self.config(UNQLITE_CONFIG_MAX_PAGE_CACHE, npages)

    # Key/Value interface.
    def store(self, key, value):
        key, value = str(key), str(value)
        return handle_return_value(unqlite_kv_store(
            self._unqlite,
            key,
            -1,
            value,
            len(value)))

    def store_fmt(self, key, value, *params):
        return handle_return_value(_unqlite_lib.unqlite_kv_store_fmt(
            self._unqlite,
            str(key),
            -1,
            value,
            *params))

    def store_file(self, key, filename):
        ptr = c_void_p()
        size = unqlite_int64(os.stat(filename).st_size)
        handle_return_value(unqlite_util_load_mmaped_file(
            filename,
            byref(ptr),
            byref(size)))
        try:
            return handle_return_value(unqlite_kv_store(
                self._unqlite,
                key,
                -1,
                ptr,
                size))
        finally:
            unqlite_util_release_mmaped_file(ptr, size)

    def append(self, key, value):
        key, value = str(key), str(value)
        handle_return_value(unqlite_kv_append(
            self._unqlite,
            key,
            -1,
            value,
            len(value)))

    def exists(self, key):
        length = unqlite_int64(0)
        key = str(key)
        return unqlite_kv_fetch(
            self._unqlite,
            key,
            -1,
            None,
            byref(length)) == UNQLITE_OK

    def fetch(self, key, buf_size=DEFAULT_BUFFER_SIZE):
        key = str(key)
        buf = create_string_buffer(buf_size)
        nbytes = unqlite_int64(buf_size)
        rc = unqlite_kv_fetch(
            self._unqlite,
            key,
            -1,
            byref(buf),
            byref(nbytes))
        if rc == UNQLITE_OK:
            return buf.raw[:nbytes.value]
        elif rc == UNQLITE_NOTFOUND:
            raise KeyError(key)
        else:
            handle_return_value(rc)

    def fetch_cb(self, key, callback):
        if not hasattr(callback, '_c_callback'):
            callback = kv_callback(callback)
        rc = unqlite_kv_fetch_callback(
            self._unqlite,
            key,
            -1,
            callback._c_callback,
            None)
        if rc == UNQLITE_NOTFOUND:
            raise KeyError(key)
        return handle_return_value(rc)

    def delete(self, key):
        key = str(key)
        rc = unqlite_kv_delete(
            self._unqlite,
            key,
            -1)
        if rc == UNQLITE_NOTFOUND:
            raise KeyError(key)
        return handle_return_value(rc)

    __getitem__ = fetch
    __setitem__ = store
    __delitem__ = delete
    __contains__ = exists

    def VM(self):
        return VM(self._unqlite)

    @contextmanager
    def compile_script(self, code):
        vm = self.VM()
        vm.compile(code)
        with vm:
            yield vm

    @contextmanager
    def compile_file(self, filename):
        vm = self.VM()
        vm.compile_file(filename)
        with vm:
            yield vm

    def collection(self, name):
        return Collection(self, name)

    # Utilities.
    def random_string(self, nbytes):
        buf = create_string_buffer(nbytes)
        handle_return_value(unqlite_util_random_string(
            self._unqlite,
            addressof(buf),
            nbytes))
        return buf.raw[:nbytes]

    def random_number(self):
        return unqlite_util_random_num(self._unqlite)

    # Library info.
    @property
    def lib_version(self):
        return str(unqlite_lib_version())

    @property
    def lib_signature(self):
        return str(unqlite_lib_signature())

    @property
    def lib_ident(self):
        return str(unqlite_lib_ident())

    @property
    def lib_is_threadsafe(self):
        return bool(unqlite_lib_is_threadsafe())

    # Transaction helpers. Currently these do not seem to work, not sure why.
    def begin(self):
        return handle_return_value(unqlite_begin(self._unqlite))

    def commit(self):
        return handle_return_value(unqlite_commit(self._unqlite))

    def rollback(self):
        return handle_return_value(unqlite_rollback(self._unqlite))

    def transaction(self):
        return transaction(self)

    def commit_on_success(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            with self.transaction():
                return fn(*args, **kwargs)
        return wrapper

    # Cursor helpers.
    def cursor(self):
        return Cursor(self._unqlite)

    def __iter__(self):
        return DBCursorIterator(self.cursor())

    def range(self, start_key, end_key, include_end_key=True):
        with self.cursor() as cursor:
            cursor.seek(start_key)
            for item in cursor.fetch_until(end_key, include_end_key):
                yield item


class Cursor(object):
    def __init__(self, unqlite):
        self._unqlite = unqlite
        self._cursor = POINTER(unqlite_kv_cursor)()
        handle_return_value(unqlite_kv_cursor_init(
            self._unqlite,
            byref(self._cursor)))

    def close(self):
        handle_return_value(unqlite_kv_cursor_release(
            self._unqlite,
            self._cursor))
        self._cursor = None

    def seek(self, key, flags=UNQLITE_CURSOR_MATCH_EXACT):
        handle_return_value(unqlite_kv_cursor_seek(
            self._cursor,
            key,
            -1,
            flags))

    def first(self):
        return unqlite_kv_cursor_first_entry(self._cursor)

    def last(self):
        return unqlite_kv_cursor_last_entry(self._cursor)

    def is_valid(self):
        return bool(unqlite_kv_cursor_valid_entry(self._cursor))

    def next(self):
        return handle_return_value(unqlite_kv_cursor_next_entry(self._cursor))

    def previous(self):
        return handle_return_value(unqlite_kv_cursor_prev_entry(self._cursor))

    def reset(self):
        return handle_return_value(unqlite_kv_cursor_reset(self._cursor))

    def delete(self):
        return handle_return_value(unqlite_kv_cursor_delete_entry(
            self._cursor))

    def _kv_cursor_value(self, unqlite_fn, bufsize):
        buf = create_string_buffer(bufsize)
        buflen = unqlite_int64(bufsize)
        handle_return_value(unqlite_fn(
            self._cursor,
            byref(buf),
            byref(buflen)))
        return buf.raw[:buflen.value]

    def key(self, bufsize=DEFAULT_BUFFER_SIZE):
        return self._kv_cursor_value(unqlite_kv_cursor_key, bufsize)

    def value(self, bufsize=DEFAULT_BUFFER_SIZE):
        return self._kv_cursor_value(unqlite_kv_cursor_data, bufsize)

    def _make_callback(self, unqlite_fn, user_fn):
        if not hasattr(user_fn, '_c_callback'):
            user_fn = kv_callback(user_fn)
        handle_return_value(unqlite_fn(
            self._cursor,
            user_fn._c_callback,
            None))
        return user_fn

    def key_callback(self, fn):
        return self._make_callback(unqlite_kv_cursor_key_callback, fn)

    def value_callback(self, fn):
        return self._make_callback(unqlite_kv_cursor_data_callback, fn)

    def __iter__(self):
        return CursorIterator(self)

    def fetch_count(self, ct):
        return CursorIterator(self, ct)

    def fetch_until(self, stop_key, include_stop_key=True):
        # Yield rows until `key` is reached.
        for key, value in self:
            if key == stop_key:
                if include_stop_key:
                    yield (key, value)
                raise StopIteration
            else:
                yield (key, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._cursor is not None:
            self.close()


class CursorIterator(object):
    def __init__(self, cursor, ct=None):
        self._cursor = cursor
        self._ct = ct or -1

    def next(self):
        if self._cursor.is_valid() and self._ct != 0:
            res = (self._cursor.key(), self._cursor.value())
            self._cursor.next()
            self._ct -= 1
            return res
        else:
            raise StopIteration

    def __iter__(self):
        return self


class DBCursorIterator(CursorIterator):
    def next(self):
        try:
            return super(DBCursorIterator, self).next()
        except StopIteration:
            self._cursor.close()
            raise


class _ValueBase(object):
    def _set_value(self, ptr, python_value):
        if isinstance(python_value, basestring):
            unqlite_value_string(ptr, python_value, -1)
        elif isinstance(python_value, (list, tuple)):
            for item in python_value:
                item_ptr = self.create_value(item)
                handle_return_value(unqlite_array_add_elem(
                    ptr,
                    None,  # automatically assign key.
                    item_ptr))
                self._release_value(item_ptr)
        elif isinstance(python_value, dict):
            for key, value in python_value.items():
                item_ptr = self.create_value(value)
                handle_return_value(unqlite_array_add_strkey_elem(
                    ptr,
                    key,
                    item_ptr))
                self._release_value(item_ptr)
        elif isinstance(python_value, (int, long)):
            unqlite_value_int(ptr, python_value)
        elif isinstance(python_value, bool):
            unqlite_value_bool(ptr, python_value)
        elif isinstance(python_value, float):
            unqlite_value_double(ptr, python_value)
        else:
            unqlite_value_null(ptr)

    def create_value(self, value):
        if isinstance(value, (list, tuple, dict)):
            ptr = self._create_array()
        else:
            ptr = self._create_scalar()
        self._set_value(ptr, value)
        return ptr

    def _release_value(self, ptr):
        raise NotImplementedError

    def _create_scalar(self):
        raise NotImplementedError

    def _create_array(self):
        raise NotImplementedError


class VM(_ValueBase):
    def __init__(self, unqlite):
        self._unqlite = unqlite
        self._vm = None
        self._ff_registry = {}

    def compile(self, code):
        self._vm = POINTER(unqlite_vm)()
        return handle_return_value(unqlite_compile(
            self._unqlite,
            code,
            -1,
            byref(self._vm)))

    def compile_file(self, filename):
        self._vm = POINTER(unqlite_vm)()
        return handle_return_value(unqlite_compile_file(
            self._unqlite,
            filename,
            byref(self._vm)))

    def close(self):
        for name in self._ff_registry.keys():
            self.delete_foreign_function(name)
        res = handle_return_value(unqlite_vm_release(self._vm))
        self._vm = None
        return res

    def reset(self):
        return handle_return_value(unqlite_vm_reset(self._unqlite))

    def execute(self):
        return handle_return_value(unqlite_vm_exec(self._vm))

    def config(self, verb, *args):
        return handle_return_value(_unqlite_lib.unqlite_vm_config(
            self._vm,
            verb,
            *args))

    def _create_scalar(self):
        return unqlite_vm_new_scalar(self._vm)

    def _create_array(self):
        return unqlite_vm_new_array(self._vm)

    def _release_value(self, ptr):
        handle_return_value(unqlite_vm_release_value(self._vm, ptr))

    def set_value(self, name, value):
        ptr = self.create_value(value)
        self.config(UNQLITE_VM_CONFIG_CREATE_VAR, name, ptr)
        self._release_value(ptr)

    def extract(self, name):
        ptr = unqlite_vm_extract_variable(self._vm, name)
        try:
            return _convert_value(ptr)
        finally:
            self._release_value(ptr)

    def foreign_function(self, name, user_data=None):
        def _decorator(fn):
            c_callback, inner = wrap_foreign_function(fn)
            self._ff_registry[name] = c_callback
            unqlite_create_function(
                self._vm,
                name,
                c_callback,
                user_data or '')
            return inner
        return _decorator

    def delete_foreign_function(self, name):
        if name in self._ff_registry:
            del self._ff_registry[name]
        handle_return_value(unqlite_delete_function(self._vm, name))

    def __getitem__(self, name):
        return self.extract(name)

    def __setitem__(self, name, value):
        self.set_value(name, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._vm is not None:
            self.close()


class _Context(_ValueBase):
    def __init__(self, context):
        self._context = context

    def _create_scalar(self):
        return unqlite_context_new_scalar(self._context)

    def _create_array(self):
        return unqlite_context_new_array(self._context)

    def _release_value(self, ptr):
        unqlite_context_release_value(self._context, ptr)

    def push_result(self, value):
        ptr = self.create_value(value)
        unqlite_result_value(self._context, ptr)
        self._release_value(ptr)


class Collection(object):
    def __init__(self, unqlite, name):
        self.unqlite = unqlite
        self.name = name

    @contextmanager
    def _execute(self, script, **kwargs):
        with self.unqlite.compile_script(script) as vm:
            vm['collection'] = self.name
            for key, value in kwargs.items():
                vm[key] = value
            vm.execute()
            yield vm

    def _simple_execute(self, script, **kwargs):
        with self._execute(script, **kwargs) as vm:
            return vm['ret']

    def all(self):
        return self._simple_execute('$ret = db_fetch_all($collection);')

    def filter(self, filter_fn):
        script = '$ret = db_fetch_all($collection, _filter_func);'
        with self.unqlite.compile_script(script) as vm:
            @vm.foreign_function('_filter_func')
            def _filter_fn(context, obj):
                return filter_fn(obj)
            vm['collection'] = self.name
            vm.execute()
            ret = vm['ret']
            vm.delete_foreign_function('_filter_func')

        return ret

    def create(self):
        script = 'if (!db_exists($collection)) { db_create($collection); }'
        with self._execute(script) as vm:
            pass

    def drop(self):
        script = 'if (db_exists($collection)) { db_drop($collection); }'
        with self._execute(script) as vm:
            pass

    def exists(self):
        return self._simple_execute('$ret = db_exists($collection);')

    def last_record_id(self):
        return self._simple_execute('$ret = db_last_record_id($collection);')

    def current_record_id(self):
        return self._simple_execute(
            '$ret = db_current_record_id($collection);')

    def reset_cursor(self):
        with self._execute('db_reset_record_cursor($collection);'):
            pass

    def __len__(self):
        return self._simple_execute('$ret = db_total_records($collection);')

    def delete(self, record_id):
        script = '$ret = db_drop_record($collection, $record_id);'
        return self._simple_execute(script, record_id=record_id)

    def fetch(self, record_id):
        script = '$ret = db_fetch_by_id($collection, $record_id);'
        return self._simple_execute(script, record_id=record_id)

    def store(self, record):
        script = '$ret = db_store($collection, $record);'
        return self._simple_execute(script, record=record)

    def fetch_current(self):
        return self._simple_execute('$ret = db_fetch($collection);')

    __delitem__ = delete
    __getitem__ = fetch

    def error_log(self):
        return self._simple_execute('$ret = db_errlog();')


class transaction(object):
    def __init__(self, unqlite):
        self.unqlite = unqlite

    def __enter__(self):
        self.unqlite.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.unqlite.rollback()
        else:
            try:
                self.unqlite.commit()
            except:
                self.unqlite.rollback()
                raise
